"""Camera class, representing a Logi Circle device"""
# coding: utf-8
# vim:sw=4:ts=4:et:
import logging
import os
from datetime import datetime, timedelta
from subprocess import CalledProcessError
from tempfile import NamedTemporaryFile
import pytz
from aiohttp.client_exceptions import ClientResponseError
from .const import (
    PROTOCOL, ACCESSORIES_ENDPOINT, ACTIVITIES_ENDPOINT, IMAGES_ENDPOINT, JPEG_CONTENT_TYPE, FEATURES,
    MODEL_GEN_1, MODEL_GEN_2, MODEL_GEN_1_NAME, MODEL_GEN_2_NAME, MODEL_GEN_UNKNOWN_NAME,
    MODEL_GEN_1_MOUNT, MODEL_GEN_2_MOUNT_WIRED, MODEL_GEN_2_MOUNT_WIRELESS,
    UPDATE_REQUEST_THROTTLE)
from .activity import Activity
from .live_stream import LiveStream
from .utils import (_stream_to_file, _write_to_file, _delete_quietly, _get_file_duration,
                    _get_first_frame_from_video, _truncate_video, requires_ffmpeg)
from .exception import UnexpectedContentType

_LOGGER = logging.getLogger(__name__)


class Camera():
    """Generic implementation for Logi Circle camera."""

    def __init__(self, logi, camera):
        """Initialise Logi Camera object."""
        self._logi = logi
        self._attrs = {}
        # Internal prop to determine when the next update is allowed.
        self._next_update_time = datetime.utcnow()

        self._set_attributes(camera)

    def _set_attributes(self, camera):
        config = None

        # Mandatory attributes
        try:
            self._attrs['id'] = camera['accessoryId']
            self._attrs['name'] = camera['name']
            self._attrs['is_connected'] = camera['isConnected']
            self._attrs['node_id'] = camera['nodeId']
            config = camera['configuration']
        except ValueError:
            _LOGGER.error(
                'Camera could not be initialised, API did not return one or more required properties.')
            raise

        # Optional attributes
        self._attrs['streaming_mode'] = config.get(
            'streamingMode', 'off') != 'off'
        self._attrs['timezone'] = config.get('timeZone', 'UTC')
        self._attrs['battery_level'] = config.get('batteryLevel', None)
        self._attrs['is_charging'] = config.get('batteryCharging', None)
        self._attrs['battery_saving'] = config.get('saveBattery', None)
        self._attrs['model'] = config.get('modelNumber', 'Unknown')
        self._attrs['wifi_signal_strength'] = config.get(
            'wifiSignalStrength', None)
        self._attrs['firmware'] = config.get('firmwareVersion', None)
        self._attrs['ip_address'] = config.get('ipAddress', None)
        self._attrs['mac_address'] = config.get('macAddress', None)
        self._attrs['wifi_ssid'] = config.get('wifiSsid', None)
        self._attrs['microphone_on'] = config.get('microphoneOn', False)
        self._attrs['microphone_gain'] = config.get('microphoneGain', None)
        self._attrs['speaker_on'] = config.get('speakerOn', False)
        self._attrs['speaker_volume'] = config.get('speakerVolume', None)
        self._attrs['led'] = config.get('ledEnabled', False)
        self._attrs['privacy_mode'] = config.get('privacyMode', False)
        self._attrs['plan_name'] = camera.get('planName', False)

        # These sensors don't appear to be present on any devices sold in the market today.
        if config.get('humidityIsAvailable', False):
            self._attrs['humidity'] = config.get('humidity', None)
        else:
            self._attrs['humidity'] = None
        if config.get('temperatureIsAvailable', False):
            self._attrs['temperature'] = config.get('temperature', None)
        else:
            self._attrs['temperature'] = None

        self._local_tz = pytz.timezone(self._attrs['timezone'])

    async def update(self, force=False):
        """Poll API for changes to camera properties."""
        _LOGGER.debug('Updating properties for camera %s', self.name)

        if force is True or datetime.utcnow() >= self._next_update_time:
            url = '%s/%s' % (ACCESSORIES_ENDPOINT, self.id)
            camera = await self._logi._fetch(
                url=url, method='GET')

            self._set_attributes(camera)
            self._next_update_time = datetime.utcnow(
            ) + timedelta(seconds=UPDATE_REQUEST_THROTTLE)
        else:
            _LOGGER.debug('Request to update ignored, requested too soon after previous update. Throttle interval is %s.',
                          UPDATE_REQUEST_THROTTLE)

    def supported_features(self):
        """Returns an array of supported sensors for this camera."""
        return FEATURES[self.mount]

    def supports_feature(self, feature):
        """Returns a bool indicating whether a given sensor is implemented for this camera."""
        return True if feature in self.supported_features() else False

    async def query_activity_history(self, property_filter=None, date_filter=None, date_operator='<=', limit=100):
        """Filter the activity history, returning Activity objects for any matching result."""

        if limit > 100:
            # Logi Circle API rejects requests where the limit exceeds 100, so we'll guard for that here.
            raise ValueError(
                'Limit may not exceed 100 due to API restrictions.')
        if date_filter is not None and not isinstance(date_filter, datetime):
            raise ValueError('date_filter must be a datetime object.')

        # Base payload object
        payload = {
            'limit': limit,
            'scanDirectionNewer': True
        }
        if date_filter:
            # Date filters are expressed using the same format for activity ID keys (YYYYMMDD"T"HHMMSSZ).
            # Let's convert our date_filter to match.

            # If timezone unaware, assume it's local to the camera's timezone.
            date_filter_tz = date_filter.tzinfo or self._local_tz

            # Activity ID keys are always expressed in UTC, so cast to UTC first.
            utc_date_filter = date_filter.replace(
                tzinfo=date_filter_tz).astimezone(pytz.utc)
            payload['startActivityId'] = utc_date_filter.strftime(
                '%Y%m%dT%H%M%SZ')

            payload['operator'] = date_operator

        if property_filter:
            payload['filter'] = property_filter

        url = '%s/%s%s' % (ACCESSORIES_ENDPOINT, self.id, ACTIVITIES_ENDPOINT)

        raw_activitites = await self._logi._fetch(
            url=url, method='POST', request_body=payload)

        activities = []
        for raw_activity in raw_activitites['activities']:
            activity = Activity(camera=self, activity=raw_activity,
                                url=url, local_tz=self._local_tz, logi=self._logi)
            activities.append(activity)

        return activities

    @property
    def live_stream(self):
        """Return LiveStream object."""
        return LiveStream(self, self._logi)

    @property
    async def last_activity(self):
        """Returns the most recent activity as an Activity object."""

        payload = {
            'limit': 1,
            'operator': '<=',
            'scanDirectionNewer': True
        }
        url = '%s/%s%s' % (ACCESSORIES_ENDPOINT, self.id, ACTIVITIES_ENDPOINT)

        raw_activity = await self._logi._fetch(
            url=url, method='POST', request_body=payload)

        try:
            unwrapped_activity = raw_activity['activities'][0]
            return Activity(camera=self, activity=unwrapped_activity, url=url, local_tz=self._local_tz, logi=self._logi)
        except IndexError:
            # If there's no activity history for this camera at all.
            return None

    @requires_ffmpeg
    async def record_livestream(self, filename, duration):
        """Downloads the live stream into a specific file for a specific duration"""
        live_stream = self.live_stream

        # Unfortunately, the Logi API does not correctly report the segment length.
        # Worse still, segments lengths can fluctuate between requests.
        # We need to check the duration of the file each iteration and abort
        # the loop once the desired duration is reached.
        downloaded_duration = 0
        max_duration = duration.total_seconds() * 1000
        failed_segment_requests = 0
        while downloaded_duration < max_duration:
            # Bail out of the loop if too many segments fail
            if failed_segment_requests > 3:
                raise RuntimeError(
                    'Live stream video length could not be evaluated after 3 attempts.')

            await live_stream.get_segment(filename=filename, append=True)
            try:
                file_duration = _get_file_duration(filename)
                downloaded_duration = file_duration
                _LOGGER.debug('Downloaded %sms of %sms of live stream to %s.',
                              file_duration, max_duration, filename)
            except (CalledProcessError, ValueError):
                # Stream may not be ready yet.
                _LOGGER.warning(
                    'Could not evaluate length of live stream segment.')
                failed_segment_requests += 1

        # Truncate video to match desired duration
        temp_file = '%s.tmp' % (filename)
        _LOGGER.debug('Trimming trailing %sms from video, using %s as temp file.',
                      downloaded_duration - max_duration, temp_file)
        try:
            _truncate_video(filename, temp_file, duration.total_seconds())
            os.remove(filename)
            os.rename(temp_file, filename)
        except OSError:
            _delete_quietly(temp_file)
            raise

    @property
    def snapshot_url(self):
        """Returns the URL that provides a near realtime JPEG snapshot of what the camera sees."""
        url = '%s://%s/api%s/%s%s' % (PROTOCOL, self.node_id,
                                      ACCESSORIES_ENDPOINT, self.id, IMAGES_ENDPOINT)
        return url

    async def get_snapshot_image(self, filename=None):
        """Downloads a near realtime JPEG snapshot for this camera."""

        # Update camera before retrieving snapshot as node ID changes frequently
        await self.update()
        url = self.snapshot_url

        image = await self._logi._fetch(
            url=url, relative_to_api_root=False, raw=True)

        if image.content_type == JPEG_CONTENT_TYPE:
            # Got an image!

            if filename:
                # Stream to file
                await _stream_to_file(image.content, filename)
                image.close()
            else:
                # Return binary object
                content = await image.read()
                image.close()
                return content
        else:
            _LOGGER.error('Expected content-type %s, got %s when retrieving still image for camera %s',
                          JPEG_CONTENT_TYPE, image.content_type, self.name)
            raise UnexpectedContentType()

    @requires_ffmpeg
    async def get_livestream_image(self, filename=None):
        """Downloads a realtime JPEG snapshot for this camera, generated from the camera's livestream."""

        live_stream = self.live_stream

        # Get an empty temp file to store our live stream segment
        segment_temp_file = NamedTemporaryFile(delete=False)
        segment_temp_file.close()
        segment_temp_file_path = segment_temp_file.name

        image = None

        try:
            # Download first segment from livestream
            await live_stream.get_segment(filename=segment_temp_file_path)

            # Get JPEG from FFMPEG
            image = _get_first_frame_from_video(segment_temp_file_path)

            # Delete temp file
            os.remove(segment_temp_file_path)
        except (CalledProcessError, ValueError):
            _delete_quietly(segment_temp_file_path)
            raise

        if filename:
            _write_to_file(data=image, filename=filename)
        else:
            return image

    async def set_streaming_mode(self, status):
        """Sets streaming mode for this camera."""
        if not isinstance(status, bool):
            raise ValueError('bool expected for status argument')
        if status is True:
            # Gen 1 cameras expect 'on', gen 2 cameras expect 'onAlert'
            if self.model == MODEL_GEN_1:
                await self._set_config(prop='streamingMode', internal_prop='streaming_mode',
                                       value='on', internal_value=True, value_type=str)
            else:
                await self._set_config(prop='streamingMode', internal_prop='streaming_mode',
                                       value='onAlert', internal_value=True, value_type=str)
        else:
            await self._set_config(prop='streamingMode', internal_prop='streaming_mode',
                                   value='off', internal_value=False, value_type=str)

    async def set_microphone(self, status=None, gain=None):
        """Sets microphone status and/or gain."""
        if status is not None:
            await self._set_config(prop='microphoneOn', internal_prop='microphone_on',
                                   value=status, value_type=bool)
        if gain is not None:
            await self._set_config(prop='microphoneGain', internal_prop='microphone_gain',
                                   value=gain, value_type=int)

    async def set_speaker(self, status=None, volume=None):
        """Sets speaker status and/or volume."""
        if status is not None:
            await self._set_config(prop='speakerOn', internal_prop='speaker_on', value=status, value_type=bool)
        if volume is not None:
            await self._set_config(prop='speakerVolume', internal_prop='speaker_volume', value=volume, value_type=int)

    async def set_led(self, status):
        """Sets LED on or off."""
        await self._set_config(prop='ledEnabled', internal_prop='led', value=status, value_type=bool)

    async def set_privacy_mode(self, status):
        """Sets privacy mode on or off."""
        await self._set_config(prop='privacyMode', internal_prop='privacy_mode', value=status, value_type=bool)

    async def set_battery_saving_mode(self, status):
        """Sets the battery saving mode on or off."""
        await self._set_config(prop='saveBattery', internal_prop='battery_saving', value=status, value_type=bool)

    async def _set_config(self, prop, internal_prop, value, value_type, internal_value=None):
        """Internal method for updating the camera's configuration."""
        if not isinstance(value, value_type):
            raise ValueError('%s expected for status argument' % (value_type))

        url = '%s/%s' % (ACCESSORIES_ENDPOINT, self.id)
        payload = {prop: value}

        _LOGGER.debug('Setting %s to %s', prop, str(value))

        try:
            req = await self._logi._fetch(
                url=url, method='PUT', request_body=payload, raw=True)
            req.close()

            # Update camera props to reflect change
            self._attrs[internal_prop] = internal_value if internal_value is not None else value
            _LOGGER.debug('Successfully set %s to %s', prop,
                          str(value))
        except ClientResponseError as error:
            _LOGGER.error(
                'Status code %s returned when updating %s to %s', error.code, prop, str(value))
            raise

    @property
    def id(self):
        """Return device ID."""
        return self._attrs.get('id')

    @property
    def node_id(self):
        """Return device node ID."""
        return self._attrs.get('node_id')

    @property
    def name(self):
        """Return device name."""
        return self._attrs.get('name')

    @property
    def timezone(self):
        """Return timezone offset."""
        return self._attrs.get('timezone')

    @property
    def is_connected(self):
        """Return bool indicating whether device is online and can accept commands (hard "on")."""
        return self._attrs.get('is_connected')

    @property
    def streaming_mode(self):
        """Return streaming mode for camera (soft "on")."""
        return self._attrs.get('streaming_mode')

    @property
    def battery_level(self):
        """Return battery level (integer between 0-100)."""
        battery_level = self._attrs.get('battery_level')
        try:
            return int(battery_level)
        except ValueError:
            return battery_level

    @property
    def battery_saving(self):
        """Return whether battery saving mode is activated."""
        return self._attrs.get('battery_saving')

    @property
    def is_charging(self):
        """Return bool indicating whether the device is currently connected to power."""
        return self._attrs.get('is_charging')

    @property
    def model(self):
        """Return model number."""
        return self._attrs.get('model')

    @property
    def model_generation(self):
        """Return product type, derived from other camera properties."""
        if self.model == MODEL_GEN_1:
            return MODEL_GEN_1_NAME
        if self.model == MODEL_GEN_2:
            return MODEL_GEN_2_NAME
        return MODEL_GEN_UNKNOWN_NAME

    @property
    def mount(self):
        """Returns the camera mount type."""
        if self.model == MODEL_GEN_1:
            return MODEL_GEN_1_MOUNT
        if self.model == MODEL_GEN_2:
            if isinstance(self.battery_level, int):
                if self.battery_level < 0:
                    return MODEL_GEN_2_MOUNT_WIRED
                return MODEL_GEN_2_MOUNT_WIRELESS
            return MODEL_GEN_2_MOUNT_WIRED
        return MODEL_GEN_UNKNOWN_NAME

    @property
    def firmware(self):
        """Return firmware version."""
        return self._attrs.get('firmware')

    @property
    def signal_strength_percentage(self):
        """Return signal strength between 0-100 (0 = bad, 100 = excellent)."""
        return self._attrs.get('wifi_signal_strength')

    @property
    def signal_strength_category(self):
        """Interpret signal strength value and return a friendly categorisation."""
        signal_strength = self._attrs.get('wifi_signal_strength', 0)
        if signal_strength > 80:
            return 'Excellent'
        if signal_strength > 60:
            return 'Good'
        if signal_strength > 40:
            return 'Fair'
        if signal_strength > 20:
            return 'Poor'
        return 'Bad'

    @property
    def temperature(self):
        """Return temperature (returns None if not supported by device)."""
        return self._attrs.get('temperature')

    @property
    def humidity(self):
        """Return relative humidity (returns None if not supported by device)."""
        return self._attrs.get('humidity')

    @property
    def ip_address(self):
        """Return local IP address for camera."""
        return self._attrs.get('ip_address')

    @property
    def mac_address(self):
        """Return MAC address for camera's WiFi interface."""
        return self._attrs.get('mac_address')

    @property
    def wifi_ssid(self):
        """Return WiFi SSID name the camera last connected with."""
        return self._attrs.get('wifi_ssid')

    @property
    def microphone_on(self):
        """Return bool indicating whether microphone is enabled."""
        return self._attrs.get('microphone_on')

    @property
    def microphone_gain(self):
        """Return microphone gain using absolute scale (1-100)."""
        gain = self._attrs.get('microphone_gain')
        try:
            return int(gain)
        except ValueError:
            return gain

    @property
    def speaker_on(self):
        """Return bool indicating whether speaker is currently enabled."""
        return self._attrs.get('speaker_on')

    @property
    def speaker_volume(self):
        """Return speaker volume using absolute scale (1-100)."""
        volume = self._attrs.get('speaker_volume')
        try:
            return int(volume)
        except ValueError:
            return volume

    @property
    def led_on(self):
        """Return bool indicating whether LED is enabled."""
        return self._attrs.get('led')

    @property
    def privacy_mode(self):
        """Return bool indicating whether privacy mode is enabled (ie. no activities recorded)."""
        return self._attrs.get('privacy_mode')

    @property
    def plan_name(self):
        """Return plan/subscription product assigned to camera (free tier, paid tier, etc)."""
        return self._attrs.get('plan_name')



"""Activity class, represents activity observed by your camera (maximum 3 minutes)"""
# coding: utf-8
# vim:sw=4:ts=4:et:
from datetime import datetime, timedelta
import logging
import pytz
from .const import (ISO8601_FORMAT_MASK, VIDEO_CONTENT_TYPE, API_URI)
from .utils import _stream_to_file
from .exception import UnexpectedContentType

_LOGGER = logging.getLogger(__name__)


class Activity():
    """Generic implementation for a Logi Circle activity."""

    def __init__(self, camera, activity, url, local_tz, logi):
        """Initialize Activity object."""
        self._camera = camera
        self._logi = logi
        self._attrs = {}
        self._local_tz = local_tz
        self._url = url
        self._set_attributes(activity)

    def _set_attributes(self, activity):
        self._attrs['activity_id'] = activity['activityId']
        self._attrs['relevance_level'] = activity['relevanceLevel']

        raw_start_time = activity['startTime']
        raw_end_time = activity['endTime']
        raw_duration = activity['playbackDuration']

        self._attrs['start_time_utc'] = datetime.strptime(
            raw_start_time, ISO8601_FORMAT_MASK)
        self._attrs['end_time_utc'] = datetime.strptime(
            raw_end_time, ISO8601_FORMAT_MASK)

        self._attrs['start_time'] = self._attrs['start_time_utc'].replace(
            tzinfo=pytz.utc).astimezone(self._local_tz)
        self._attrs['end_time'] = self._attrs['end_time_utc'].replace(
            tzinfo=pytz.utc).astimezone(self._local_tz)

        self._attrs['duration'] = timedelta(milliseconds=raw_duration)

    @property
    def download_url(self):
        """Returns the download URL for the current activity."""
        return '%s%s/%s/mp4' % (API_URI, self._url, self.activity_id)

    async def download(self, filename=None):
        """Download the activity as an MP4, optionally saving to disk."""
        url = self.download_url

        video = await self._logi._fetch(url=url, method='GET', raw=True, relative_to_api_root=False)

        if video.content_type == VIDEO_CONTENT_TYPE:
            # Got a video!
            if filename:
                # Stream to file
                await _stream_to_file(video.content, filename)
                video.close()
            else:
                # Return binary object
                content = await video.read()
                video.close()
                return content
        else:
            _LOGGER.error('Expected content-type %s, got %s when retrieving activity video.',
                          VIDEO_CONTENT_TYPE, video.content_type)
            raise UnexpectedContentType()

    @property
    def activity_id(self):
        """Return activity ID."""
        return self._attrs['activity_id']

    @property
    def start_time(self):
        """Return start time as datetime object, local to the camera's timezone."""
        return self._attrs['start_time']

    @property
    def end_time(self):
        """Return end time as datetime object, local to the camera's timezone."""
        return self._attrs['end_time']

    @property
    def start_time_utc(self):
        """Return start time as datetime object in the UTC timezone."""
        return self._attrs['start_time_utc']

    @property
    def end_time_utc(self):
        """Return end time as datetime object in the UTC timezone."""
        return self._attrs['end_time_utc']

    @property
    def duration(self):
        """Return activity duration as a timedelta object."""
        return self._attrs['duration']



"""Utilities library shared by the Logi, Camera and Activity classes."""
# coding: utf-8
# vim:sw=4:ts=4:et:
import os
import functools
import subprocess
import logging
try:
    import cPickle as pickle
except ImportError:
    import pickle
from logi_circle.const import (
    CACHE_ATTRS, COOKIE_NAME)
from .exception import BadSession

_LOGGER = logging.getLogger(__name__)


def _get_session_cookie(cookie_jar):
    """Iterates through the session's AbstractCookieJar and returns the cookie relevant to Logi API sessions."""
    for cookie in cookie_jar:
        if cookie.key == COOKIE_NAME:
            return cookie
    raise BadSession()


async def _handle_response(request, raw):
    """Generic response handler, intended to transparently handle session expiry."""
    if request.status == 401:
        # Session has likely expired. Re-authentiate.
        raise BadSession()

    # Throw unhandled error if request failed for any other reason
    request.raise_for_status()

    if raw:
        return request
    else:
        resp = await request.json()
        request.close()
        return resp


async def _stream_to_file(stream, filename, open_mode='wb'):
    """Stream aiohttp response to file."""
    with open(filename, open_mode) as file_handle:
        while True:
            chunk = await stream.read(1024)
            if not chunk:
                break
            file_handle.write(chunk)


def _write_to_file(data, filename, open_mode='wb'):
    """Write binary object directly to file."""
    with open(filename, open_mode) as file_handle:
        file_handle.write(data)


def _delete_quietly(filename):
    """Deletes a file which may or may not exist."""
    try:
        os.remove(filename)
    except OSError:
        pass


def _get_file_duration(file):
    """Get the duration in milliseconds of a video using ffprobe."""
    output = subprocess.check_output(
        ['ffprobe', '-i', file, '-show_entries',
         'format=duration', '-v', 'quiet', '-of', 'csv=p=0'],
        stderr=subprocess.STDOUT
    )
    return int(float(output) * 1000)


def _truncate_video(input_file, output_file, seconds):
    """Truncates a video based on the input duration"""
    subprocess.check_call(
        ['ffmpeg', '-y', '-i', input_file, '-t', str(seconds),
         '-f', 'mp4', '-c', 'copy', output_file],
        stderr=subprocess.DEVNULL
    )


def _get_first_frame_from_video(file):
    """Returns the first frame of a video as a bytes object using ffmpeg."""
    output = subprocess.check_output(
        ['ffmpeg', '-i', file, '-vf',
         'select=eq(n\\,0)', '-q:v', '3', '-f', 'singlejpeg', '-'],
        stderr=subprocess.DEVNULL
    )
    return output


def _ffmpeg_installed():
    """Returns a bool indicating whether ffmpeg is installed."""
    try:
        subprocess.check_call(["ffmpeg", "-version"],
                              stdout=subprocess.DEVNULL)
        return True
    except OSError:
        _LOGGER.warning(
            'ffmpeg is not installed! Not all API methods will function.')
        return False


_FFMPEG_INSTALLED = _ffmpeg_installed()


def requires_ffmpeg(func):
    """Decorator to guard against ffmpeg use if it's not installed."""
    @functools.wraps(func)
    def wrapped_decorator(*args, **kwargs):
        if not _FFMPEG_INSTALLED:
            raise RuntimeError(
                'This method requires ffmpeg to be installed and available from the current execution context.')
        else:
            return func(*args, **kwargs)
    return wrapped_decorator


def _write_to_file(data, filename):
    """Write bytes data to file."""
    with open(filename, 'wb') as file_handle:
        file_handle.write(data)


def _clean_cache(filename):
    """Remove filename if pickle version mismatch."""
    if os.path.isfile(filename):
        os.remove(filename)

    # Initialise cache since file was removed
    initial_cache_data = CACHE_ATTRS
    _save_cache(initial_cache_data, filename)
    return initial_cache_data


def _exists_cache(filename):
    """Check if filename exists and if is pickle object."""
    return bool(os.path.isfile(filename))


def _save_cache(data, filename):
    """Dump data into a pickle file."""
    with open(filename, 'wb') as pickle_db:
        pickle.dump(data, pickle_db)
    return True


def _read_cache(filename):
    """Read data from a pickle file."""
    try:
        if os.path.isfile(filename):
            data = pickle.load(open(filename, 'rb'))

            if data.keys() != CACHE_ATTRS.keys():
                raise EOFError
            return data

    except (EOFError, ValueError):
        pass
    return _clean_cache(filename)



__all__ = [
    'UnexpectedContentType', 'BadSession', 'NoSession', 'BadCache'
]


class UnexpectedContentType(Exception):
    """When the wrong content type is received for a request"""


class BadSession(Exception):
    """Request made on a session that's no longer valid"""


class NoSession(Exception):
    """No session active"""


class BadCache(Exception):
    """Cached credentials incomplete, corrupt or do not apply to current user"""


class BadLogin(Exception):
    """Username or password rejected"""



"""Live Stream class"""
# coding: utf-8
# vim:sw=4:ts=4:et:
import logging
import asyncio
from datetime import datetime, timedelta
import xml.etree.ElementTree as ET
from .const import (PROTOCOL, ACCESSORIES_ENDPOINT,
                    LIVESTREAM_ENDPOINT, LIVESTREAM_XMLNS)
from .utils import _stream_to_file, _write_to_file

_LOGGER = logging.getLogger(__name__)


class LiveStream():
    """Logi Circle DASH client."""

    def __init__(self, camera, logi):
        """Initialize LiveStream object."""
        self._camera = camera
        self._logi = logi
        self._initialisation_file = None
        self._index = None
        self._initialised = False
        self._mpd_data = {}
        self._next_segment_time = None
        self._last_filename = None

    def _get_mpd_url(self):
        """Returns the URL for the MPD file"""
        return '%s://%s/api%s/%s%s' % (PROTOCOL, self._camera.node_id,
                                       ACCESSORIES_ENDPOINT, self._camera.id, LIVESTREAM_ENDPOINT)

    async def _get_init_mp4(self, base_url, header_url):
        """Downloads the initialisation file and returns a bytes object"""
        url = '%s%s' % (base_url, header_url)

        header = await self._logi._fetch(
            url=url, relative_to_api_root=False, raw=True)

        header_data = await header.read()
        header.close()
        return header_data

    async def _get_mpd(self):
        """Gets the MPD XML and extracts the data required to download segments"""

        # Force an update to get the latest node ID
        await self._camera.update(force=True)

        # Get MPD XML and save to raw_xml var
        url = self._get_mpd_url()

        xml_req = await self._logi._fetch(
            url=url, relative_to_api_root=False, raw=True)
        raw_xml = await xml_req.read()
        xml_req.close()

        # Extract data from MPD XML
        xml = ET.fromstring(raw_xml)
        stream_config = {}
        stream_config['base_url'] = xml.find(
            './{%s}BaseURL' % (LIVESTREAM_XMLNS)).text
        stream_config['header_url'] = xml.find('.//{%s}SegmentTemplate' %
                                               (LIVESTREAM_XMLNS)).get('initialization')
        stream_config['stream_filename_template'] = xml.find(
            './/{%s}SegmentTemplate' % (LIVESTREAM_XMLNS)).get('media')
        stream_config['start_index'] = int(xml.find('.//{%s}SegmentTemplate' %
                                                    (LIVESTREAM_XMLNS)).get('startNumber'))
        stream_config['segment_length'] = int(xml.find('.//{%s}SegmentTemplate' %
                                                       (LIVESTREAM_XMLNS)).get('duration'))
        return stream_config

    def _build_mp4(self, segment_file):
        """Concatenates the initialisation data and segment file to return a playable MP4"""

        return self._initialisation_file + segment_file

    def _set_next_segment_time(self):
        """Sets the time that the next segment should be ready to download"""
        duration = self._mpd_data['segment_length']
        self._next_segment_time = datetime.now() + timedelta(milliseconds=duration)

    def _get_time_before_next_segment(self):
        """Time before the next segment is available, in seconds"""
        delay = self._next_segment_time - datetime.now()
        delay_in_seconds = delay.total_seconds()
        return 0 if delay_in_seconds < 0 else delay_in_seconds

    def _get_segment_url(self):
        """Builds the URL to get the next video segment"""
        base_url = self._mpd_data['base_url']
        stream_filename_template = self._mpd_data['stream_filename_template']

        file_name = stream_filename_template.replace(
            '$Number$', str(self._index))
        return '%s%s' % (base_url, file_name)

    async def get_segment(self, filename=None, append=True):
        """Returns the current segment video from the live stream"""
        # Initialise if required
        if self._initialised is False:
            await self._initialise()

        # Get current wait time and set timer for next download
        wait_time = self._get_time_before_next_segment()
        self._set_next_segment_time()

        _LOGGER.debug(
            'Sleeping for %s seconds before grabbing next live stream segment.', wait_time)
        # And sleep, if needed.
        if wait_time > 0:
            await asyncio.sleep(wait_time)

        # Get segment data
        url = self._get_segment_url()
        segment = await self._logi._fetch(
            url=url, relative_to_api_root=False, raw=True)

        # Increment segment counter
        self._index += 1

        if filename:
            if filename == self._last_filename and append:
                # Append to existing file
                _LOGGER.debug(
                    'Appending video segment to %s', filename)
                await _stream_to_file(stream=segment.content, filename=filename, open_mode='ab')
                segment.close()
            else:
                # Write init header and segment (init a new file)
                _LOGGER.debug(
                    'Writing init header and segment to file %s', filename)
                _write_to_file(self._initialisation_file, filename)
                await _stream_to_file(stream=segment.content, filename=filename, open_mode='ab')
                segment.close()
            self._last_filename = filename
        else:
            # Return binary object
            content = await segment.read()
            segment.close()
            return self._build_mp4(content)

    async def _initialise(self):
        """Sets up the live stream so that it's ready to output video data"""

        # Get stream config and cache header
        self._mpd_data = await self._get_mpd()
        self._initialisation_file = await self._get_init_mp4(self._mpd_data['base_url'], self._mpd_data['header_url'])
        self._index = self._mpd_data['start_index']
        self._initialised = True

        # Delay stream until one segment is ready
        self._set_next_segment_time()



"""Logi API"""
# coding: utf-8
# vim:sw=4:ts=4:et:
import sys
import logging
import json
import aiohttp


from .utils import _get_session_cookie, _handle_response, _exists_cache, _save_cache, _read_cache, _clean_cache
from .const import (
    API_URI, AUTH_ENDPOINT, CACHE_ATTRS, CACHE_FILE, CAMERAS_ENDPOINT, COOKIE_NAME, VALIDATE_ENDPOINT, HEADERS)
from .camera import Camera
from .exception import BadSession, NoSession, BadCache, BadLogin

_LOGGER = logging.getLogger(__name__)


class Logi():
    """A Python abstraction object to Logi Circle cameras."""

    def __init__(self, username, password, reuse_session=True, cache_file=CACHE_FILE):
        self.is_connected = False
        self.params = None
        self.username = username
        self.password = password

        self._session = None
        self._cache = CACHE_ATTRS
        self._cache_file = cache_file
        self._reuse_session = reuse_session

    async def login(self):
        """Create a session with the Logi Circle API"""
        # Close current session before creating a new one
        if isinstance(self._session, aiohttp.ClientSession):
            await self._session.close()

        if self._reuse_session:
            try:
                # Restore cached cookie and validate.
                await self._restore_cached_session()
            except BadCache:
                # Fall back to authentication if restoring the cached session fails.
                await self._authenticate()
        else:
            await self._authenticate()

    async def logout(self):
        """Close the session with the Logi Circle API"""
        _LOGGER.debug('Closing session for %s.', self.username)
        if isinstance(self._session, aiohttp.ClientSession):
            await self._session.close()
            self._session = None
            self.is_connected = False
        else:
            raise NoSession()

    async def _authenticate(self):
        """Authenticate user with the Logi Circle API."""
        url = API_URI + AUTH_ENDPOINT
        login_payload = {'email': self.username, 'password': self.password}

        _LOGGER.debug("POSTing login payload to %s", url)

        session = aiohttp.ClientSession()
        async with session.post(url, json=login_payload, headers=HEADERS) as req:
            # Handle failed authentication due to incorrect user/pass
            if req.status == 401:
                raise BadLogin(
                    'Username or password provided is incorrect. Could not authenticate.')

            # Throw error if authentication failed for any reason (connection issues, outage, etc)
            req.raise_for_status()

            self._session = session
            self.is_connected = True
            _LOGGER.info("Logged in as %s.", self.username)

            # Cache account and cookie for later use if reuse_session is true
            if self._reuse_session:
                try:
                    cookie = _get_session_cookie(session.cookie_jar)
                    _LOGGER.debug("Persisting session cookie %s", cookie)
                    self._cache['account'] = self.username
                    self._cache['cookie'] = cookie.value
                    _save_cache(self._cache, self._cache_file)
                except BadSession:
                    _LOGGER.error(
                        "Logi API authenticated successfully, but the session cookie wasn't returned.")
                    raise

    async def _restore_cached_session(self):
        """Retrieved cached cookie and validate session."""
        if _exists_cache(self._cache_file):
            _LOGGER.debug("Restoring cookie from cache.")

            self._cache = _read_cache(self._cache_file)

            if (self._cache['account'] is None) or (self._cache['cookie'] is None):
                # Cache is missing one or required values.
                _LOGGER.debug('Cache appears corrupt. Re-authenticating.')
                raise BadCache()
            elif self._cache['account'] != self.username:
                # Cache does not apply to this user.
                _LOGGER.debug(
                    'Cached credentials are for a different user. Re-authenticating.')
                raise BadCache()
            else:
                cookies = {COOKIE_NAME: self._cache['cookie']}

                session = aiohttp.ClientSession(cookies=cookies)
                async with session.get(API_URI + VALIDATE_ENDPOINT, headers=HEADERS) as req:
                    if req.status >= 300:
                        # Cookie has probably expired, reauthenticate.
                        raise BadCache()
                    else:
                        _LOGGER.info(
                            "Restored cached session for %s.", self.username)
                        self._session = session
                        self.is_connected = True

        else:
            raise BadCache()

    async def _fetch(self,
                     url,
                     method='GET',
                     params=None,
                     request_body=None,
                     relative_to_api_root=True,
                     raw=False,
                     _reattempt=False):
        """Query data from the Logi Circle API."""

        if self._session is None:
            await self.login()

        resolved_url = (API_URI + url if relative_to_api_root else url)
        _LOGGER.debug("Fetching %s (%s)", resolved_url, method)

        req = None

        # Perform request
        if method == 'GET':
            req = await self._session.get(resolved_url, headers=HEADERS, params=params)
        elif method == 'POST':
            req = await self._session.post(resolved_url, headers=HEADERS, params=params, json=request_body)
        elif method == 'PUT':
            req = await self._session.put(resolved_url, headers=HEADERS, params=params, json=request_body)
        else:
            raise ValueError('Method %s not supported.' % (method))

        _LOGGER.debug('Request %s (%s) returned %s',
                      resolved_url, method, req.status)

        # Handle response
        try:
            return await _handle_response(request=req, raw=raw)
        except BadSession:
            if _reattempt:
                # Welp, session still bad even after reauthenticating.
                _LOGGER.error(
                    'Session expired and a new session could not be established.')
                raise
            else:
                # Assume session expiry. Re-authenticate and try again.
                _LOGGER.debug(
                    'Session appears to have expired. Re-authenticating.')
                await self.login()
                return await self._fetch(url=url,
                                         method=method,
                                         params=params,
                                         request_body=request_body,
                                         relative_to_api_root=relative_to_api_root,
                                         raw=raw,
                                         _reattempt=True)

    @property
    def cookie(self):
        """Returns a cookie string that can be used for requests external to this library"""
        if self._session is None:
            return False
        return '%s=%s' % (COOKIE_NAME, _get_session_cookie(self._session.cookie_jar).value)

    @property
    async def cameras(self):
        """ Return all cameras. """
        cameras = []
        raw_cameras = await self._fetch(CAMERAS_ENDPOINT)

        for camera in raw_cameras:
            cameras.append(Camera(self, camera))

        return cameras



# coding: utf-8
# vim:sw=4:ts=4:et:
"""Constants"""
import os

# Requests and cached session properties
HEADERS = {
    'User-Agent': 'iOSClient/3.4.5.31',
    'Accept-Encoding': 'gzip, deflate, br',
    'Accept-Language': 'en-GB,en-US;q=0.9,en;q=0.8',
    'Connection': 'keep-alive',
    'Origin': 'https://circle.logi.com',
    'Accept': '*/*',
    'Referer': 'https://circle.logi.com/',
    'Pragma': 'no-cache'
}
CACHE_ATTRS = {'cookie': None, 'account': None}
try:
    CACHE_FILE = os.path.join(os.getenv("HOME"), '.logi_circle-session.cache')
except (AttributeError, TypeError):
    CACHE_FILE = os.path.join('.', '.logi_circle-session.cache')
COOKIE_NAME = 'prod_session'
UPDATE_REQUEST_THROTTLE = 5  # seconds

# Date formats
# Yes, we're hard-coding the timezone. "%z" is a py37 only feature,
# and we know the Logi API always returns dates in UTC.
ISO8601_FORMAT_MASK = '%Y-%m-%dT%H:%M:%SZ'

# API endpoints
PROTOCOL = 'https'
API_HOST = 'video.logi.com'
API_URI = '%s://%s/api' % (PROTOCOL, API_HOST)
AUTH_ENDPOINT = '/accounts/authorization'
CAMERAS_ENDPOINT = '/accessories'
IMAGES_ENDPOINT = '/image'
LIVESTREAM_ENDPOINT = '/mpd'
ACTIVITIES_ENDPOINT = '/activities'
ACCESSORIES_ENDPOINT = '/accessories'
VALIDATE_ENDPOINT = '/accounts/self'

# Misc
JPEG_CONTENT_TYPE = 'image/jpeg'
VIDEO_CONTENT_TYPE = 'application/octet-stream'
LIVESTREAM_XMLNS = 'urn:mpeg:dash:schema:mpd:2011'

# Model to product mapping
MODEL_GEN_1 = 'A1533'
MODEL_GEN_2 = 'V-R0008'
MODEL_GEN_1_NAME = '1st generation'
MODEL_GEN_2_NAME = '2nd generation'
MODEL_GEN_UNKNOWN_NAME = 'Unknown'
MODEL_GEN_1_MOUNT = 'Charging ring'
MODEL_GEN_2_MOUNT_WIRED = 'Wired'
MODEL_GEN_2_MOUNT_WIRELESS = 'Wireless'

# Feature mapping
FEATURES = {
    MODEL_GEN_1_MOUNT: ['is_charging', 'battery_level', 'last_activity_time', 'privacy_mode',
                        'signal_strength_percentage', 'signal_strength_category', 'speaker_volume', 'streaming_mode'],
    MODEL_GEN_2_MOUNT_WIRED: ['last_activity_time', 'privacy_mode', 'signal_strength_percentage',
                              'signal_strength_category', 'speaker_volume', 'streaming_mode'],
    MODEL_GEN_2_MOUNT_WIRELESS: ['is_charging', 'battery_level', 'last_activity_time', 'privacy_mode',
                                 'signal_strength_percentage', 'signal_strength_category', 'speaker_volume', 'streaming_mode'],
    MODEL_GEN_UNKNOWN_NAME: ['last_activity_time']
}



# coding=utf-8
from setuptools import setup


def readme():
    with open('README.md', encoding='utf-8') as desc:
        return desc.read()


setup(
    name='logi_circle',
    packages=['logi_circle'],
    version='0.1.8',
    description='A Python library to communicate with Logi Circle cameras',
    long_description=readme(),
    long_description_content_type='text/markdown',
    author='Evan Bruhn',
    author_email='evan.bruhn@gmail.com',
    url='https://github.com/evanjd/python-logi-circle',
    license='MIT',
    include_package_data=True,
    install_requires=['aiohttp', 'pytz'],
    test_suite='tests',
    keywords=[
        'logi',
        'logi circle',
        'logitech'
        'home automation',
    ],
    classifiers=[
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Framework :: AsyncIO',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Home Automation',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)
